/*
 * Copyright 2006-2021 The MZmine Development Team
 *
 * This file is part of MZmine.
 *
 * MZmine is free software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * MZmine is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with MZmine; if not,
 * write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

package io.github.mzmine.modules.dataprocessing.id_isotopepeakscanner;

import io.github.mzmine.datamodel.features.FeatureList;
import io.github.mzmine.datamodel.features.FeatureListRow;
import java.util.ArrayList;
import java.util.Set;
import java.util.TreeMap;


/**
 * This class can be used to handle feature lists by ID rather than row indices. Set up via the
 * constructor by passing a feature list or via setUp(PeakList) Rows can also be added manually, for
 * example if you want to create a result feature list that does not contain duplicates. Since this
 * class uses a tree map the results will be in order and duplicates will be overwritten, which is
 * why it has the method containsID to check.
 *
 * @author Steffen Heuckeroth steffen.heuckeroth@gmx.de / s_heuc03@uni-muenster.de
 *
 */
public class PeakListHandler {

  private TreeMap<Integer, FeatureListRow> map;

  public PeakListHandler() {
    map = new TreeMap<Integer, FeatureListRow>();
  }

  public PeakListHandler(FeatureList pL) {
    map = new TreeMap<Integer, FeatureListRow>();
    setUp(pL);
  }

  /**
   * use this if you want to manage an existing PeakList
   * 
   * @param pL the feature list you want to manage
   */
  public void setUp(FeatureList pL) {
    for (FeatureListRow row : pL.getRows()) {
      map.put(row.getID(), row);
    }
  }

  /**
   * Manually add a PeakListRow
   * 
   * @param row row to be added
   */
  public void addRow(FeatureListRow row) {
    map.put(row.getID(), row);
  }

  /**
   * 
   * @return number of rows handled with plh
   */
  public int size() {
    return map.size();
  }

  /**
   * 
   * @param ID ID to check for
   * @return true if contained, false if not
   */
  public boolean containsID(int ID) {
    return map.containsKey(ID);
  }

  /**
   * 
   * @return ArrayList<Integer> of all IDs of the feature list rows
   */
  public ArrayList<Integer> getAllKeys() {
    Set<Integer> set = map.keySet();
    ArrayList<Integer> list = new ArrayList<Integer>(set);

    return list;
  }

  /**
   * 
   * @param ID ID of the row you want
   * @return Row with specified ID
   */
  public FeatureListRow getRowByID(int ID) {
    return map.get(ID);
  }

  /**
   * 
   * @param ID integer array of IDs
   * @return all rows with specified ids
   */
  public FeatureListRow[] getRowsByID(int ID[]) {
    FeatureListRow[] rows = new FeatureListRow[ID.length];

    for (int i = 0; i < ID.length; i++)
      rows[i] = map.get(ID[i]);

    return rows;
  }
}
